// DFT.CPP Implementation of DFT
// version 1.0
// (c) 2005 Chad Berchek

#include <cmath>
#include "wavetypes.h"
#include "dft.h"
using namespace std;

DFTobject::DFTobject(const sampleT *dataPtr) : sampleData(dataPtr)
{
	frequency = 0;
	len = 0;
	samplingRate = 1;
	invalidate();
}

void DFTobject::setFreq(frequencyT newFreq) {
	if (newFreq != frequency) {
		frequency = newFreq;
		invalidate();
	}
}

void DFTobject::setLen(sampleCountT n) {
	if (n != len) {
		len = n;
		invalidate();
	}
}

void DFTobject::setSamplingRate(samplingRateT newSamplingRate) {
	if (newSamplingRate != samplingRate) {
		samplingRate = newSamplingRate;
		invalidate();
	}
}

void DFTobject::setFreqAndLen(frequencyT freq, sampleCountT n) {
	if (freq != frequency || n != len) {
		frequency = freq;
		len = n;
		invalidate();
	}
}

void DFTobject::setNewData(const sampleT* newData) {
	if (newData != sampleData) {
		sampleData = newData;
		invalidate();
	}
}

void DFTobject::invalidate() {
	raw.isCurrent = false;
	ideal.isCurrent = false;
	dual.isCurrent = false;
}
		
amplitudeT DFTobject::getDualDftAmplitude() {
	if (!dual.isCurrent) DualDft();
	return dual.amplitude;
}

phaseT DFTobject::getDualDftPhase() {
	if (!dual.isCurrent) DualDft();
	return dual.phase;
}

amplitudeT DFTobject::getRawAmplitude() {
	if (!raw.isCurrent) Dft();
	return raw.amplitude;
}

phaseT DFTobject::getRawPhase() {
	if (!raw.isCurrent) Dft();
	return raw.phase;
}

amplitudeT DFTobject::getIdealAmplitude() {
	if (!ideal.isCurrent) IdealDft();
	return ideal.amplitude;
}

phaseT DFTobject::getIdealPhase() {
	if (!ideal.isCurrent) IdealDft();
	return ideal.phase;
}

void DFTobject::IdealDft() {
	if (!raw.isCurrent) Dft();

	double arg, generateArg;
	double cosPart = 0, sinPart = 0;
		
	for (sampleCountT k = 0; k < len; k++) {
		arg =  2. * M_PI * frequency / samplingRate * k;
		generateArg = arg + raw.phase;
		sinPart += sin(generateArg) * sin(arg);
		cosPart += sin(generateArg) * cos(arg);
	}

	ideal.amplitude = 2. * sqrt( sinPart*sinPart + cosPart*cosPart ) / len;
	ideal.phase = atan2(cosPart, sinPart);
	ideal.isCurrent = true;
}

void DFTobject::Dft() {
	if (sampleData != 0) {
		double cosPart = 0, sinPart = 0;
		double arg;
		for (sampleCountT k = 0; k < len; k++) {
			arg =  2. * M_PI * frequency / samplingRate * k;
			sinPart += double(sampleData[k]) * sin(arg);
			cosPart += double(sampleData[k]) * cos(arg);
		}
		raw.amplitude = 2. * sqrt( sinPart*sinPart + cosPart*cosPart ) / len;
		if (sinPart == 0 && cosPart == 0) raw.phase = 0;
		else raw.phase = atan2(cosPart, sinPart);
	}
	else {
		raw.amplitude = 0;
		raw.phase = 0;
	}

	raw.isCurrent = true;
}


void DFTobject::DualDft() {
	if (!raw.isCurrent) Dft();
	if (!ideal.isCurrent) IdealDft();
	
	dual.amplitude = raw.amplitude / ideal.amplitude;
	dual.phase = raw.phase;
	
	dual.isCurrent = true;
}
