// Sweep.cpp  Sweep class implementation
// version 1.0
// (c) 2005 Chad Berchek

#include "sweep.h"
#include "wavetypes.h"
#include "miscmath.h"
#include <cmath>
#include <cstdio>
#include <cstring>

using namespace std;

const char* Sweep::ID = "SWEEPPARAMS";

void Sweep::setStartFreq(frequencyT sf) {
	if (!locked && sf > 0) {
		startFreq = sf;
		if (startFreq > endFreq) endFreq = startFreq;
		reset();
	}
}

void Sweep::setSweepType(sweepType st) {
	if (!locked) {
		currentSweepType = st;
		reset();
	}
}

void Sweep::setFreqStep(frequencyT fs) {
	if (!locked && fs > 0) {
		freqStep = fs;
		reset();
	}
}

void Sweep::setEndFreq(frequencyT ef) {
	if (!locked && ef > 0) {
		endFreq = ef;
		if (endFreq < startFreq) startFreq = endFreq;
	}
}

void Sweep::setSyncVolume(percentVolumeT vol) {
	if (!locked) {
		if (vol < 0) syncVolume = 0;
		else if (vol > 100) syncVolume = 100;
		else syncVolume = vol;
	}
}

void Sweep::setSyncThreshold(percentVolumeT st) {
	if (!locked) {
		if (st < 0) syncThreshold = 0;
		else if (st > 100) syncThreshold = 100;
		else syncThreshold = st;
	}
}

double Sweep::getPercentDone() const {
	return 100. * double(toneCounter) / getTotalNumTones();
}

size_t Sweep::getTotalNumTones() const {
	if (currentSweepType == LINEAR) {
		return size_t( (endFreq - startFreq) / freqStep + 1 );
	}
	else { // else octave sweep
		return size_t( log(endFreq / startFreq) / log(2.) * freqStep + 1 );
	}
}

bool Sweep::writeTo(FILE *outFile) const {
	if (fwrite(ID, IDlen, 1, outFile) != 1) return false;
	if (fwrite(this, sizeof(Sweep), 1, outFile) != 1) return false;
	return true;
}

bool Sweep::readFrom(FILE *inFile) {
	if (locked) return false;

	char readID[IDlen];
	if (fread(readID, IDlen, 1, inFile) != 1) return false;
	if (strncmp(ID, readID, IDlen) != 0) return false; // Not a SweepParams file

	if (fread(this, sizeof(Sweep), 1, inFile) != 1) return false;

	reset();
	return true;
}

Sweep::Sweep(FILE *file) {
	locked = false;
	if (!readFrom(file)) initialize();
}

void Sweep::initialize() {
	locked = false;
	startFreq = 1;
	endFreq = 1;
	freqStep = 1;
	toneDuration = 0;
	syncDuration = 0;
	settlingTime = 0;
	skipBeginning = 0;
	maxTimeDrift = 0;
	samplingRate = 1.;
	maxExtraSamples = 0;
	maxFractionalCycle = 1.;
	currentFreq = 1.;
	tonesPerSync = 0;
	toneCounter = 0;
	currentToneDurationS = 0;
	currentSweepType = LINEAR;
	syncVolume = 0;
	syncThreshold = 0;
	minNumCycles = 0;
}

void Sweep::reset() {
	toneCounter = 0;
	setCurrentFreq(calcNthFreq(toneCounter));
}

bool Sweep::needsSync() const {
	if (tonesPerSync == 0) {
		if (toneCounter == 0) return true;
		else return false;
	}
	
	return ( (toneCounter % tonesPerSync) == 0 );
}

frequencyT Sweep::calcNthFreq(long n) const {
	if (currentSweepType == LINEAR) {
		return startFreq + n * freqStep;
	}
	else { // currentSweepType == OCTAVE
		return startFreq * pow(2., n / freqStep);
	}
}

bool Sweep::goToNextFreq() {
	toneCounter++;
	frequencyT newFreq = calcNthFreq(toneCounter);

	if (newFreq > endFreq) {
      toneCounter--;
   	return false;
	}

	setCurrentFreq(newFreq);
	
	return true;
}

void Sweep::setCurrentFreq(frequencyT newFreq) {
	currentFreq = newFreq;
	currentToneDurationS = getNumSamples(currentFreq);
}	

double Sweep::getFractionOfCycle(double cycles) const {
	double fracCycle = fpart(cycles);
	double from1 = 1. - fracCycle;
	double x;
	
	x = (fracCycle <= from1) ? fracCycle : from1;
	return x;
}

double Sweep::getNumSamples(double freq) const {
	double errorFreq = 2. * freq;
	if (errorFreq > samplingRate / 2.) errorFreq = samplingRate - errorFreq;
	if (errorFreq == 0.) return toneDuration * samplingRate;
	
	double eqFreq = errorFreq / 2.;  // Equivalent frequency - frequency
               	// which has the same error frequency as the actual frequency
	double numCycles = ceil( toneDuration * eqFreq );
	if (numCycles < 10) numCycles = 10;
	double realNumSamples;
	double bestFraction = 1.;
	double bestNumSamples = numCycles / eqFreq * samplingRate;
	double fractionalCycle;
	double iNumSamples;
	
	double extraCycles = 0;
	do {
		realNumSamples = double(numCycles + extraCycles) / eqFreq * samplingRate;
		iNumSamples = fround(realNumSamples);
      fractionalCycle = getFractionOfCycle(double(iNumSamples) / samplingRate * eqFreq);
		if (fractionalCycle < bestFraction) {
			bestFraction = fractionalCycle;
			bestNumSamples = realNumSamples;
		}
		if (bestFraction <= maxFractionalCycle) break;
		extraCycles += 1.;
	} while ((extraCycles / eqFreq * samplingRate) < maxExtraSamples);
	
	return fround(bestNumSamples);
}

double Sweep::getPhaseAtSample(long n) const {
	double samplesPerCycle = double(samplingRate) / currentFreq;
	double fractionOfCycle = fmod(n, samplesPerCycle) / samplesPerCycle;
	double phase = fractionOfCycle * 2. * M_PI;
	return phase;
}

// Returns the number of samples needed to pad the end of the current tone
sampleCountT Sweep::getEndPadDurationS() const {
	double currentToneCycles = double(getCurrentToneDurationS()) / samplingRate * currentFreq;
	// The number of cycles of the tone itself is included so that the whole wave will, including
	// the end pad, will come out with a whole number of cycles. The number of samples in the
	// tone itself must be subtracted off at the end.
	double numCycles = maxTimeDrift * currentFreq + currentToneCycles;
	numCycles = ceil(numCycles);
	return sampleCountT(numCycles / currentFreq * samplingRate) - getCurrentToneDurationS();
}

// Returns the number of samples needed for settling time after the current tone
sampleCountT Sweep::getSettlingTimeS() const {
	double numCycles = settlingTime * currentFreq;
	numCycles = ceil(numCycles);
	return sampleCountT(double(numCycles) / currentFreq * samplingRate);
}
