// Sweep.h  Sweep class
// version 1.0
// (c) 2005 Chad Berchek

#ifndef Sweep_h
#define Sweep_h

#include "wavetypes.h"
#include <cstdio>
using namespace std;

class Sweep {
	public:	
		Sweep() { initialize(); }
		Sweep(FILE *file);
		
		enum sweepType { LINEAR, OCTAVE };

		void setStartFreq(frequencyT sf);
		void setEndFreq(frequencyT ef);
		void setFreqStep(frequencyT fs);
		void setToneDuration(timeT td)			{ if (!locked && td >= 0) toneDuration = td; }
		void setMaxTimeDrift(timeT md)			{ if (!locked && md >= 0) maxTimeDrift = md; }
		void setSettlingTime(timeT st)			{ if (!locked && st >= 0) settlingTime = st; }
		void setSkipBeginning(timeT sb) 			{ if (!locked && sb >= 0) skipBeginning = sb; }
		void setSyncDuration(timeT sd) 			{ if (!locked && sd >= 0) syncDuration = sd; }
		void setSamplingRate(samplingRateT sr)	{ if (!locked) samplingRate = sr; }
		void setTonesPerSync(size_t tps)			{ if (!locked) tonesPerSync = tps; }
		void setMaxFractionalCycle(double mfc)	{ if (!locked && mfc >= 0) maxFractionalCycle = mfc; }
		void setMaxExtraSamples(sampleCountT m){ if (!locked) maxExtraSamples = m; }
		void setMinNumCycles(cycleCountT mc)	{ if (!locked) minNumCycles = mc; }
		void setSweepType(sweepType st);
		void setSyncVolume(percentVolumeT vol);
		void setSyncThreshold(percentVolumeT st);

		frequencyT getStartFreq()			const { return startFreq; }
		frequencyT getEndFreq()				const { return endFreq; }
		frequencyT getFreqStep()			const { return freqStep; }
		timeT getToneDuration()				const { return toneDuration; }
		timeT getMaxTimeDrift()				const { return maxTimeDrift; }
		timeT getSettlingTime()				const { return settlingTime; }
		timeT getSkipBeginning()			const { return skipBeginning; }
		timeT getSyncDuration()				const { return syncDuration; }
		samplingRateT getSamplingRate()	const { return samplingRate; }
		size_t getTonesPerSync()			const { return tonesPerSync; }
		double getMaxFractionalCycle()	const { return maxFractionalCycle; }
		sampleCountT getMaxExtraSamples() const { return maxExtraSamples; }
		sweepType getSweepType()			const { return currentSweepType; }
		percentVolumeT getSyncVolume()	const { return syncVolume; }
		percentVolumeT getSyncThreshold() const { return syncThreshold; }
		cycleCountT getMinNumCycles()		const { return minNumCycles; }
		

		bool goToNextFreq();
		bool needsSync()						const;
		size_t getToneCounter()				const { return toneCounter; }
		double getPhaseAtSample(long n)	const;
		frequencyT getCurrentFreq()		const { return currentFreq; }
		sampleCountT getSyncDurationS()	const { return syncDuration * samplingRate; }
		sampleCountT getSettlingTimeS()	const;
		sampleCountT getSkipBeginningS()	const { return skipBeginning * samplingRate; }
		sampleCountT getPreSyncBreakS()	const { return maxTimeDrift * samplingRate; }
		sampleCountT getPreSyncSettlingS() const { return settlingTime * samplingRate; }
		sampleCountT getEndPadDurationS() const;
		sampleCountT getCurrentToneDurationS()	const { return currentToneDurationS; }
		double getFractionOfCycle(double cycles) const;
		double getNumSamples(double freq) const;		
		frequencyT calcNthFreq(long n)	const;
		double getPercentDone()				const;
		size_t getTotalNumTones()			const;

		void lock() { locked = true; }
		void reset();
		
		bool writeTo(FILE *outFile) const;
		bool readFrom(FILE *inFile);
		
	private:	
	
		bool locked;
		frequencyT startFreq;
		frequencyT endFreq;
		frequencyT freqStep;
		timeT toneDuration;
		timeT syncDuration;
		timeT settlingTime;
		timeT maxTimeDrift;
		timeT skipBeginning;
		samplingRateT samplingRate;
		sampleCountT maxExtraSamples;
		double maxFractionalCycle;
		frequencyT currentFreq;
		size_t tonesPerSync;
		size_t toneCounter;
		sampleCountT currentToneDurationS;
		sweepType currentSweepType;
		percentVolumeT syncVolume;
		percentVolumeT syncThreshold;
		cycleCountT minNumCycles;
		static const char* ID;
		static const size_t IDlen = 11;
		
		void initialize();
		void setCurrentFreq(frequencyT newFreq);
};

#endif
