// WaveGen.cpp  Main WaveGen program - generates test tones and sweep parameters file
// version 1.0
// (c) 2005 Chad Berchek

// *** Un-comment the line below to build for Windows platforms.
//#define USE_MS_WIN_DIALOGS

#define _WIN32_WINNT 0x0400 // Keep this line if building for Windows 95/98

//////////////////////////////////////////////////////////////////////////////////////////

#include <algorithm>
#include <cstdio>
#include <cmath>
#include "wave.h"
#include "miscmath.h"
#include "wavetypes.h"
#include "sweep.h"
#include "kbinput.h"
using namespace std;

// Sweep parameters: minimums, maximums, and defaults
const double MIN_SAMPLING_RATE = 8000;
const double MAX_SAMPLING_RATE = 96000;
const double DEF_SAMPLING_RATE = 48000;
const double DEF_START_FREQ = 20;
const double MIN_FREQ = 0.1;
const double DEF_END_FREQ = 20000;
const double MIN_LINEAR_FREQ_STEP = 0.01;
const double MAX_LINEAR_FREQ_STEP = 1e10;
const double DEF_LINEAR_FREQ_STEP = 10;
const double MIN_TONES_PER_OCTAVE = .1;
const double MAX_TONES_PER_OCTAVE = 5000;
const double DEF_TONES_PER_OCTAVE = 24;
const double MIN_TONE_DURATION = 0;
const double MAX_TONE_DURATION = 1000;
const double DEF_TONE_DURATION = 0.02;
const double MAX_TONES_PER_SYNC = 1e10;
const double DEF_TONES_PER_SYNC = 0;
const double MIN_FRACTIONAL_CYCLE = 0;
const double MAX_FRACTIONAL_CYCLE = 1;
const double DEF_FRACTIONAL_CYCLE = 1;
const double MAX_EXTRA_SAMPLES = 3000;
const double MIN_EXTRA_SAMPLES = 0;
const double DEF_EXTRA_SAMPLES = 400;
const double MIN_SYNC_AMP = 10;
const double MAX_SYNC_AMP = 100;
const double DEF_SYNC_AMP = 90;
const double MIN_TIME_DRIFT = 0;
const double MAX_TIME_DRIFT = 1;
const double DEF_TIME_DRIFT = .0002;
const double MIN_SETTLING_TIME = 0;
const double MAX_SETTLING_TIME = 1;
const double DEF_SETTLING_TIME = .04;
const double DEF_SYNC_DURATION = .0002;
const double DEF_SKIP_BEGINNING = .5;
const unsigned long MIN_MIN_CYCLES = 0;
const unsigned long MAX_MIN_CYCLES = 200;
const unsigned long DEF_MIN_CYCLES = 10;
const size_t FILE_NAME_BUF_SIZE = 600;
const timeT LEADER = 3; // 3 seconds in addition to the skipBeginning value
const percentVolumeT TONE_VOLUME = 100;
const percentVolumeT DEF_SYNC_THRESHOLD = 15;

// Input user settings from keyboard and store in sweep object
void getSettings(Sweep &sweep) {
	printf("\nEnter the parameters for this frequency sweep. For any values you do not");
	printf("\ncare about, just press Enter to accept the default.\n");

	printf("\nEnter output sampling rate in Hz (between %.0f and %.0f; default=%.0f): ",
		MIN_SAMPLING_RATE, MAX_SAMPLING_RATE, DEF_SAMPLING_RATE);
	sweep.setSamplingRate( kbReadDouble( MIN_SAMPLING_RATE, MAX_SAMPLING_RATE, DEF_SAMPLING_RATE ));
	const double nyquist = sweep.getSamplingRate() / 2.;
	
	printf("\nSelect sweep type: 1) Octave or 2) Linear (default=Octave): ");
	int sweepTypeInput = kbReadDouble(1, 2, 1);
	if (sweepTypeInput == 1) sweep.setSweepType(Sweep::OCTAVE);
	else sweep.setSweepType(Sweep::LINEAR);
	
	printf("\nEnter the starting frequency of the sweep in Hz (default=%.0f): ", DEF_START_FREQ);
	sweep.setStartFreq( kbReadDouble(MIN_FREQ, nyquist, DEF_START_FREQ ));
	
	double defaultEndFreq = min(DEF_END_FREQ, nyquist);
	printf("\nEnter the end frequency of the sweep in Hz. The maximum frequency must be");
	printf("\nat least the starting frequency and less than half the sampling rate.");
	printf("\nEnter the end frequency (between %.0f and %.0f; default=%.0f): ",
 		double(sweep.getStartFreq()), nyquist, defaultEndFreq);
	sweep.setEndFreq( kbReadDouble(sweep.getStartFreq(), nyquist, defaultEndFreq) );

	if (sweep.getSweepType() == Sweep::LINEAR) {
		printf("\nEnter the frequency step in Hz from one tone to the next (default=%.0f): ", DEF_LINEAR_FREQ_STEP);
		sweep.setFreqStep( kbReadDouble(MIN_LINEAR_FREQ_STEP, MAX_LINEAR_FREQ_STEP, DEF_LINEAR_FREQ_STEP) );
	}
	else if (sweep.getSweepType() == Sweep::OCTAVE) {
		printf("\nEnter the number of tones to play per octave (default=%.0f): ", DEF_TONES_PER_OCTAVE);
		sweep.setFreqStep( kbReadDouble(MIN_TONES_PER_OCTAVE, MAX_TONES_PER_OCTAVE, DEF_TONES_PER_OCTAVE) );
	}
	
	printf("\nEnter the minimum duration for each tone in seconds (default=%.2f): ", DEF_TONE_DURATION);
	sweep.setToneDuration( kbReadDouble(MIN_TONE_DURATION, MAX_TONE_DURATION, DEF_TONE_DURATION) );

	printf("\nEnter the time between tones (in seconds) needed for each tone to stabilize.");
	printf("\n Recommended settings:");
	printf("\n  Speakers:       at least 0.040 seconds");
	printf("\n  High-quality headphones: 0.025 seconds");
	printf("\n  Electronic circuits:     0.005 seconds");
	printf("\n Enter settling time (between %.0f and %.0f; default=%.2f): ", MIN_SETTLING_TIME, MAX_SETTLING_TIME, DEF_SETTLING_TIME);
	sweep.setSettlingTime( kbReadDouble(MIN_SETTLING_TIME, MAX_SETTLING_TIME, DEF_SETTLING_TIME) );
	
	printf("\nEnter the number of tones to be played following each synchronization pulse.");
	printf("\nThe default, 0, will generate only one synchronization pulse,\nat the beginning. ");
	sweep.setTonesPerSync( kbReadDouble(0, MAX_TONES_PER_SYNC, DEF_TONES_PER_SYNC) );
		
	printf("\nDo you want to edit the advanced options? (y/n) ");
	if (kbReadYN()) {
		printf("\nEnter the amplitude of the synchronization pulse as a percent of the maximum");
		printf("\nvolume (between %.0f and %.0f; default=%.0f): ", MIN_SYNC_AMP, MAX_SYNC_AMP, DEF_SYNC_AMP);
		sweep.setSyncVolume( kbReadDouble(MIN_SYNC_AMP, MAX_SYNC_AMP, DEF_SYNC_AMP) );

		printf("\nEnter the minimum number of cycles to play of any frequency (this helps ensure");
		printf("\nthat enough data is collected for the low frequencies).");
		printf("\nMinimum number of cycles (between %u and %u; default=%u): ",
			MIN_MIN_CYCLES, MAX_MIN_CYCLES, DEF_MIN_CYCLES);
		sweep.setMinNumCycles(kbReadDouble(MIN_MIN_CYCLES, MAX_MIN_CYCLES, DEF_MIN_CYCLES));

		printf("\nEnter the maximum allowable fractional cycle. This is an error-control value.");
		printf("\nA value of 1 will work in general, but use values of 0.1 or less to reduce");
		printf("\nerror. Enter maximum fractional cycle (between %.2f and %.1f; default=%.1f): ", 
			MIN_FRACTIONAL_CYCLE, MAX_FRACTIONAL_CYCLE, DEF_FRACTIONAL_CYCLE);
		sweep.setMaxFractionalCycle( kbReadDouble(MIN_FRACTIONAL_CYCLE, MAX_FRACTIONAL_CYCLE, DEF_FRACTIONAL_CYCLE) );

		printf("\nEnter the maximum number of extra samples to be added to each tone in order");
		printf("\nto control error. Larger values will result in larger wave files but possibly");
		printf("\nhigher accuracy.");
		printf("\nEnter the maximum number of samples (between %.0f and %.0f; default=%.0f): ",
			MIN_EXTRA_SAMPLES, MAX_EXTRA_SAMPLES, DEF_EXTRA_SAMPLES);
		sweep.setMaxExtraSamples( kbReadDouble(MIN_EXTRA_SAMPLES, MAX_EXTRA_SAMPLES, DEF_EXTRA_SAMPLES) );
		
		printf("\nEnter the maximum expected drift (in seconds) due to unequal sampling rates");
		printf("\nbetween the tone source and recorder (between %.0f and %.0f; default=%.4f): ",
			MIN_TIME_DRIFT, MAX_TIME_DRIFT, DEF_TIME_DRIFT);
		sweep.setMaxTimeDrift( kbReadDouble(MIN_TIME_DRIFT, MAX_TIME_DRIFT, DEF_TIME_DRIFT) );
	} 
	else {
		sweep.setMaxExtraSamples(DEF_EXTRA_SAMPLES);
		sweep.setSyncVolume(DEF_SYNC_AMP);
		sweep.setMaxTimeDrift(DEF_TIME_DRIFT);
		sweep.setMinNumCycles(DEF_MIN_CYCLES);
		sweep.setMaxFractionalCycle(DEF_FRACTIONAL_CYCLE);
	}
	
	sweep.setSyncThreshold(DEF_SYNC_THRESHOLD);
	sweep.setSyncDuration(DEF_SYNC_DURATION);
	sweep.setSkipBeginning(DEF_SKIP_BEGINNING);
}

// Takes the percentage of maximum volume and gives the amplitude in terms of the sample values
sampleT convertPercentToAmplitude(const percentVolumeT percentVolume) {
	long amplitude;  // Long is used to avoid possible overflow due to rounding
						  // from floating pt. Range is checked when converting to short.
	amplitude = percentVolume / 100. * MAX_SAMPLE_VALUE;
	if (amplitude < -MAX_SAMPLE_VALUE) amplitude = -MAX_SAMPLE_VALUE;
	if (amplitude > MAX_SAMPLE_VALUE) amplitude = MAX_SAMPLE_VALUE;
	return amplitude;
}

#ifdef USE_MS_WIN_DIALOGS
	#include <windows.h>
	bool openOutputFiles(WaveFile &waveFile, FILE *&parameterFile) {
		const char *errorMsg = "The selected file could not be opened.";
		bool openSuccess;

		OPENFILENAME ofn = { 0 };
		char fileNameBuf[FILE_NAME_BUF_SIZE];

		// Initialize ofn
		ofn.lStructSize = sizeof(ofn);
		ofn.lpstrFile = fileNameBuf;
		ofn.nMaxFile = FILE_NAME_BUF_SIZE;
		ofn.Flags = OFN_NOREADONLYRETURN | OFN_OVERWRITEPROMPT;

		//////// Open wave file for output //////////
		fileNameBuf[0] = 0;
		ofn.lpstrTitle = "Enter a name for the output wave file";
		const char *wavFilter = "Wave Files (*.wav)\0*.wav\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = wavFilter;
		ofn.lpstrDefExt = "wav";
		do {
			BOOL success = GetSaveFileName(&ofn);
			if (!success) return false;
			openSuccess = waveFile.OpenWrite(fileNameBuf);
			if (!openSuccess) {
				MessageBox(0, errorMsg, "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!openSuccess);

		// Find position of '.' before filename extension
		long waveFileNameFullLen = strlen(fileNameBuf);
		long waveFileNameExtnPos;
		if ((ofn.nFileExtension == 0) || (ofn.nFileExtension == waveFileNameFullLen)) {
			waveFileNameExtnPos = waveFileNameFullLen;  // points to '\0'
		}
		else waveFileNameExtnPos = ofn.nFileExtension - 1;  // -1 takes care of '.'
		
		// Initialize filename for sweep parameters file to wave file name minus extension
		fileNameBuf[waveFileNameExtnPos] = 0;  // truncate extension

		//////// Open sweep parameters file for output //////////
		ofn.lpstrTitle = "Enter a name for the sweep parameters file";
		const char *swpFilter = "Sweep Parameter Files (*.swp)\0*.swp\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = swpFilter;
		ofn.lpstrDefExt = "swp";
		do {
			BOOL success = GetSaveFileName(&ofn);
			if (!success) return false;
			parameterFile = fopen(fileNameBuf, "wb");
			if (!parameterFile) {
				MessageBox(0, errorMsg, "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!parameterFile);
		
		return true;
	}
#else	
	bool openOutputFiles(WaveFile &outputWaveFile, FILE *&parameterFile) {
		char buf[FILE_NAME_BUF_SIZE];
	
		printf("\nEnter the filename for the output wave file, including the .WAV extension");
		printf("\n(or enter nothing to quit this program):\n");
	
		do {
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
		} while (!outputWaveFile.OpenWrite(buf));
		
		printf("\nEnter the filename to which the sweep's parameters will be written:\n");
		do {
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
			parameterFile = fopen(buf, "wb");
		} while (!parameterFile);
		
		return true;
	}
#endif

// Writes a sine waveform to the given WaveFile. Only 16-bit sample size is
// allowed, and only mono. Sampling rate is taken into account.
// percentVolume is the percentage of maximum amplitude, i.e., max = 100.
void writeSineWave(const frequencyT freq, const phaseT phase, const percentVolumeT percentVolume,
const sampleCountT numSamples, WaveFile &out) {
	if (out.GetBitsPerChannel() != 16) return;
	
	const double SINE_ARG = 2. * M_PI * freq / out.GetSampleRate();
	const double amplitude = convertPercentToAmplitude(percentVolume);	

	for (sampleCountT i = 0; i < numSamples; i++) {
		sampleT sample = roundToShort(amplitude * sin(SINE_ARG * i + phase));
		out.WriteSample(sample);
	}
}

// Writes a series of samples at equal level (a pulse) to the wave file
void writePulse(sampleCountT numSamples, const percentVolumeT percentVolume, WaveFile &out) {
	if (out.GetBitsPerChannel() != 16) return;

	const sampleT amplitude = convertPercentToAmplitude(percentVolume);
	// Write samples	
	for (sampleCountT i = 0; i < numSamples; i++) out.WriteSample(amplitude);
}

void generateSweep(Sweep &sweep, WaveFile &outWave) {
	// Print message that the sweep is being generated
	printf("\nGenerating tones. Percent done:\n");
	
	sweep.reset();
		
	// Write silent beginning (leader)
	writePulse(sweep.getSkipBeginningS() + LEADER * sweep.getSamplingRate(), 0, outWave);
	
	do {
		if (sweep.needsSync()) {
			writePulse(sweep.getPreSyncSettlingS(), 0, outWave);
			writePulse(sweep.getPreSyncBreakS(), 0, outWave);
			writePulse(sweep.getSyncDurationS(), sweep.getSyncVolume(), outWave);
		}			
		
		// Find the phase so that after the settling period, the phase will be 0.
		sampleCountT settlingTime = sweep.getSettlingTimeS();		
		phaseT phase = sweep.getPhaseAtSample(-settlingTime);

		writeSineWave(sweep.getCurrentFreq(), phase, TONE_VOLUME,
			settlingTime + sweep.getCurrentToneDurationS() + sweep.getEndPadDurationS(), outWave);
			
		// Update percent done display
		printf("\r%.0f%%  ", sweep.getPercentDone());
	} while (sweep.goToNextFreq());
	
	// Finish off percent done display
	printf("\r100%%  \n");
}

int main() {
	printf("\n--- WaveGen ---\n");
	printf("Fransys 1.0 / Vobarian Software / (c)2004 Chad Berchek\n");
	
	// Open output files
	WaveFile outWave;
	FILE *outParam;
	if (!openOutputFiles(outWave, outParam)) return 0;
	
	// Get sweep parameters, write to disk
	Sweep sweep;
	getSettings(sweep);
	sweep.lock();
	sweep.writeTo(outParam);
	fclose(outParam);
	
	// Prepare output wave file
 	outWave.SetupFormat();
	outWave.SetSampleRate(sweep.getSamplingRate());
	
	generateSweep(sweep, outWave);
	
	return 0;
}
