// WaveMeas.cpp  Main WaveMeas program - analyzes recorded wave files
// version 1.0
// (c) 2005 Chad Berchek

// *** Un-comment the line below to build for Windows platforms.
//#define USE_MS_WIN_DIALOGS

#define _WIN32_WINNT 0x0400 // Keep this line if building for Windows 95/98

//////////////////////////////////////////////////////////////////////////////////////////

#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <memory>
#include "sweep.h"
#include "kbinput.h"
#include "wavetypes.h"
#include "wave.h"
#include "dft.h"
using namespace std;

const size_t FILE_NAME_BUF_SIZE = 600;  // file name buffer for opening files

// structure for results from one tone measurement
struct toneResult {
	frequencyT frequency;
	amplitudeT amplitude;
	phaseT phase;
};

// Takes the percentage of maximum volume and gives the amplitude in terms of the sample values
sampleT convertPercentToAmplitude(const percentVolumeT percentVolume) {
	long amplitude;  // Long is used to avoid possible overflow due to rounding
						  // from floating pt. Range is checked when converting to short.
	amplitude = percentVolume / 100. * MAX_SAMPLE_VALUE;
	if (amplitude < -MAX_SAMPLE_VALUE) amplitude = -MAX_SAMPLE_VALUE;
	if (amplitude > MAX_SAMPLE_VALUE) amplitude = MAX_SAMPLE_VALUE;
	return amplitude;
}

// Open files using Windows dialogs or text file-name entry
#ifdef USE_MS_WIN_DIALOGS
	#include <windows.h>
	bool openFiles(WaveFile &waveFile, Sweep &sweep, FILE* &outputFile, auto_ptr<char> &waveFileName) {
		bool openSuccess;
		OPENFILENAME ofn = { 0 };
		char fileNameBuf[FILE_NAME_BUF_SIZE];

		// Initialize ofn structure
		ofn.lStructSize = sizeof(ofn);
		ofn.lpstrFile = fileNameBuf;
		ofn.nMaxFile = FILE_NAME_BUF_SIZE;
		
		//////// Open wave file for input //////////
		fileNameBuf[0] = 0;
		const char *wavFilter = "Wave Files (*.wav)\0*.wav\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = wavFilter;
		ofn.Flags = OFN_FILEMUSTEXIST;
		ofn.lpstrTitle = "Select the recorded sample";
		do {
			BOOL success = GetOpenFileName(&ofn);
			if (!success) return false;
			openSuccess = waveFile.OpenRead(fileNameBuf);
			if (!openSuccess) {
				MessageBox(0, "The selected file could not be opened as a wave file.", "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!openSuccess);
		
		// Copy input wave filename to string (without path)
		waveFileName.reset( new char[FILE_NAME_BUF_SIZE] );
		strcpy(waveFileName.get(), fileNameBuf + ofn.nFileOffset);
		// Find position of '.' before filename extension
		long waveFileNameFullLen = strlen(fileNameBuf);
		long waveFileNameExtnPos;
		if ((ofn.nFileExtension == 0) || (ofn.nFileExtension == waveFileNameFullLen)) {
			waveFileNameExtnPos = waveFileNameFullLen - ofn.nFileOffset;
		}
		else waveFileNameExtnPos = ofn.nFileExtension - 1 - ofn.nFileOffset;  // -1 takes care of '.'

		//////// Open sweep parameters file for input //////////
		const char *swpFilter = "Sweep Parameter Files (*.swp)\0*.swp\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = swpFilter;
		fileNameBuf[0] = 0;
		ofn.lpstrTitle = "Select the sweep paramters file";
		ofn.lpstrDefExt = "swp";
		do {
			BOOL success = GetOpenFileName(&ofn);
			if (!success) return false;
			FILE *sweepFile = fopen(fileNameBuf, "rb");
			if (sweepFile) {
				openSuccess = sweep.readFrom(sweepFile);
			} 
			else openSuccess = false;
			if (!openSuccess) {
				MessageBox(0, "The selected file is not a list of sweep parameters.", "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!openSuccess);

		//////// Open text file for output //////////
		const char *txtFilter = "Text Files (*.txt)\0*.txt\0All Files (*.*)\0*\0";
		ofn.lpstrFilter = txtFilter;
		// Initialize filename to wave file name minus extension
		strcpy(fileNameBuf, waveFileName.get());
		fileNameBuf[waveFileNameExtnPos] = 0;  // truncate extension
		ofn.lpstrTitle = "Enter a filename for the output";
		ofn.lpstrDefExt = "txt";
		ofn.Flags = OFN_NOREADONLYRETURN | OFN_OVERWRITEPROMPT;
		do {
			BOOL success = GetSaveFileName(&ofn);
			if (!success) return false;
			outputFile = fopen(fileNameBuf, "wt");
			if (!outputFile) {
				MessageBox(0, "The selected file could not be opened for output.", "Error", MB_OK | MB_ICONEXCLAMATION);
			}
		} while (!outputFile);
		
		return true;
	}	
#else	
	bool openFiles(WaveFile &waveFile, Sweep &sweep, FILE* &outputFile, auto_ptr<char> &waveFileName) {
		char buf[FILE_NAME_BUF_SIZE];
	
		printf("\n(Note: Enter a blank filename to quit the program.)");
		printf("\nRecorded Wave filename: ");

		// Open wave file for input
		do {
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
		} while (!waveFile.OpenRead(buf));

		// Copy input wave filename to string (without path)
		waveFileName.reset( new char[FILE_NAME_BUF_SIZE] );
		strcpy(waveFileName.get(), buf);

		printf("\nSweep parameters filename:\n");
		bool sweepReadOK = false;
		do {
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
			FILE *parameterFile = fopen(buf, "rb");
			if (parameterFile) sweepReadOK = sweep.readFrom(parameterFile);
		} while (!sweepReadOK);
		
		printf("\nOutput text filename:\n");
		do {
			kbReadString(buf, FILE_NAME_BUF_SIZE);
			if (isEmptyString(buf)) return false;
			outputFile = fopen(buf, "wt");
		} while (!outputFile);
		
		return true;
	}
#endif

bool waveFormatOK(const Sweep &sweep, const WaveFile &inWave) {
	if (inWave.GetNumChannels() != 1) {
		printf("\nError: Please only use input wave files with one channel (mono).");
		return false;
	}
	if (inWave.GetBitsPerChannel() != 16) {
		printf("\nError: Please only use input wave files with 16-bit samples.");
		return false;
	}
	if (inWave.GetSampleRate() != sweep.getSamplingRate()) {
		printf("\nError: The input wave file's sampling rate is not the same as the");
		printf("\nsampling rate of the original tone sweep. The sampling rate");
		printf("\nshould be %u Hz.", (unsigned long)(sweep.getSamplingRate()));
		return false;
	}
	return true;
}

bool synchronize(const Sweep &sweep, WaveFile &wave) {
	sampleT syncThreshold = convertPercentToAmplitude(sweep.getSyncThreshold());
	sampleT inputSample;
	
	// Find synchronization pulse
	do {
		if (!wave.ReadSample(inputSample)) return false;
	} while (abs(inputSample) < syncThreshold);
	
	// Skip rest of synchronization pulse
	if (!wave.SeekFwd( sweep.getSyncDurationS() - 1)) return false;

	return true;
}

unsigned short findMaxSample(unsigned short prevMax, sampleT *sample, sampleCountT numSamples) {
	unsigned short maxSample = prevMax;
	for (sampleCountT i = 0; i < numSamples; i++) {
		unsigned short currentSample = abs(sample[i]);
		if (currentSample > maxSample) maxSample = currentSample;
	}
	return maxSample;
}

bool analyzeSweep(Sweep &sweep, WaveFile &inWave, toneResult* result, size_t nResult) {
	bool success = true;
	unsigned short maxSample = 0;
	sampleCountT sampleArraySize = 0;
	auto_ptr<sampleT> sampleArray;
	DFTobject dft;

	dft.setSamplingRate(sweep.getSamplingRate());	
	sweep.reset();

	// Print message
	printf("\nAnalyzing input wave. Percent done:\n");

	do {
		if (sweep.needsSync()) {
			inWave.SeekFwd(sweep.getPreSyncSettlingS());
			if (!synchronize(sweep, inWave)) {
				printf("\nError: The synchronization pulse could not be found.");
				success = false;
				break;
			}
		}

		// Skip settling period
		inWave.SeekFwd(sweep.getSettlingTimeS());
		
		// Allocate memory for samples
		sampleCountT toneSamples = sweep.getCurrentToneDurationS();
		if (toneSamples > sampleArraySize) {
			sampleArraySize = toneSamples;
			sampleArray.reset(new sampleT[sampleArraySize]);
			dft.setNewData(sampleArray.get());
		}

		// Read tone samples
		if (!inWave.ReadSamples(sampleArray.get(), toneSamples)) {
			printf("\nError: Input wave file ended before expected.");
			success = false;
			break;
		}
		
		// Skip end pad
		inWave.SeekFwd(sweep.getEndPadDurationS());

		// Perform DFT analysis and store results
		dft.setFreqAndLen(sweep.getCurrentFreq(), toneSamples);
		size_t toneCount = sweep.getToneCounter();
		if (toneCount < nResult) {
			result[toneCount].frequency = sweep.getCurrentFreq();
			result[toneCount].amplitude = dft.getDualDftAmplitude();
			result[toneCount].phase = dft.getDualDftPhase();
		}
		else break; // Number of tones exceeds capacity of results array
		// Update maximum sample
		maxSample = findMaxSample(maxSample, sampleArray.get(), toneSamples);
		// Update percent done display
		printf("\r%.0f%%  ", sweep.getPercentDone());
	} while (sweep.goToNextFreq());

	// Finish off percent done display and display maximum level
	if (success) {
		printf("\r100%%  \n");
		double peakLevel = double(maxSample) / MAX_SAMPLE_VALUE * 100.;
		printf("\nPeak recording level: %.0f%%\n", peakLevel);
	}
	
	return success;
}

void listSweepParams(const Sweep &sweep, const char *waveFileName, FILE *out) {
	fprintf(out, "Input file, \"%s\"\n", waveFileName);
	fprintf(out, "Sampling rate, %u Hz\n", (unsigned long)sweep.getSamplingRate());
	fprintf(out, "Minimum tone duration, %.3f sec\n", (double)sweep.getToneDuration());
	fprintf(out, "Tones per sync pulse, %u\n", (size_t)sweep.getTonesPerSync());
	fprintf(out, "Settling time, %.3f sec\n", (double)sweep.getSettlingTime());
	fprintf(out, "Minimum cycles per tone, %u\n", (unsigned long)sweep.getMinNumCycles());
	fprintf(out, "Maximum fractional cycle, %.3f\n", (double)sweep.getMaxFractionalCycle());
	fprintf(out, "Maximum extra samples, %u\n", (unsigned long)sweep.getMaxExtraSamples());
	fprintf(out, "Maximum time drift, %.4f sec\n", (double)sweep.getMaxTimeDrift());
}

void writeResults(toneResult *results, size_t nResult, FILE *output) {
	size_t i; // General counter
	
	if (nResult < 1) return; // From now on assume there is at least one tone result

	// Iterate through all results and:
	// 1. Sum the amplitudes
	// 2. Un-alias phase
	double amplitudeSum = 0.;
	double prevPhase = results[0].phase;
	//double minTimeShift = results[0].phase / results[0].frequency; // NOTE: minTimeShfit is not
                   // actually a time; it should be divided by 2PI and then multiplied by 2PI
                   // when calculating adjustedPhaseDeg below - so this factor is just left out
	for (i = 0; i < nResult; i++) {
		amplitudeSum += results[i].amplitude;
		
		while (results[i].phase - prevPhase < -M_PI_2) results[i].phase += 2. * M_PI;
		while (results[i].phase - prevPhase >  M_PI_2) results[i].phase -= 2. * M_PI;
		prevPhase = results[i].phase;

		//double timeShift = results[i].phase / results[i].frequency;
		//if (timeShift < minTimeShift) minTimeShift = timeShift;
	}
	
	// Calculate average amplitude
	amplitudeT avgAmp = amplitudeSum / nResult;
	bool avgAmpIsZero = (avgAmp == 0.);  // avoid floating-point comparison in for loop below
	
	// Write column headers
	fprintf(output, "Frequency, Amplitude, Norm. Amp., Norm. Amp. (dB), Phase (deg)\n");

	// Write data
	for (i = 0; i < nResult; i++) {
		// write frequency and amplitude (raw) and norm. amplitude
		fprintf(output, "%.2f, %.1f, ", double(results[i].frequency),
			double(results[i].amplitude));
		if (avgAmpIsZero) fprintf(output, "undef, undef, ");
		else {
			fprintf(output, "%.5f, ", double(results[i].amplitude / avgAmp));
			// Find relative amplitude in dB and write to file
			if (results[i].amplitude > 0.) {
				double ampdb = 20. * log10(results[i].amplitude / avgAmp);
				fprintf(output, "%.3f, ", ampdb);
			}
			else fprintf(output, "-Inf, ");
		}
		// Find phase in degrees and write to file
		double adjustedPhaseDeg = results[i].phase * 180. / M_PI;
		fprintf(output, "%.2f\n", adjustedPhaseDeg);
	}
}

// Some operating systems may close the program window right after termination. This function
// gives the user a chance to view messages.
void closeWait() {
	printf("\nPress <Enter> to end the program.\n");
	kbReadEnter();
}

int main() {
	printf("\n--- WaveMeas ---\n");
	printf("Fransys 1.0 / Vobarian Software / (c)2005 Chad Berchek\n");

	// Open files
	WaveFile inWave;
	Sweep sweep;
	FILE *outputFile;
	auto_ptr<char> waveFileName;
	if (!openFiles(inWave, sweep, outputFile, waveFileName)) return 0;
	sweep.lock();

	atexit(closeWait);  // gives the user a chance to view messages
	
	// Check Wave file format
	if (!waveFormatOK(sweep, inWave)) return 0;
	
	// Write sweep paramteres to output file
	listSweepParams(sweep, waveFileName.get(), outputFile);

	// Skip a small part of the beginning of the file, since it may contain garbage which would
	// be misinterpreted as the synchronization pulse
	if (!inWave.SeekFwd( sweep.getSkipBeginningS() )) {
		printf("\nError: Input wave file ended before expected.");
		return 0;
	}

	// Allocate array for results
	size_t numTones = sweep.getTotalNumTones();
	toneResult *results = new toneResult[numTones];
	// Perform analysis
	if (!analyzeSweep(sweep, inWave, results, numTones)) return 0;
	// Write results
	writeResults(results, numTones, outputFile);
	
	return 0;
}
